/*ScianPeriodicTable.h
  Eric Pepke
  Information partially entered by Michelle Maples

	Modified June 1993 by Christina Lam to include covalent radius data
	from periodic table issued by Sargent-Welch Canada.

	"Tableau des Proprietes Periodiques des Elements"
	Sargent-Welch Scientifique Canada Limitee
	285 Garyray Drive
	Weston, Ontario M9L 1P3
	
	numero du catalogue S-18806-09
	Tous droits reserves 1979 (all rights reserved)
*/

#define MISRAD -1.0		/* extern real MISSINGDATA would be appropriate here,
								    but an array can't be initialized with variable
								    elements */

AtomInfo atomInfo[N_ATOMS] =
    {
/*     Symbol    Name             Weight AtomicRadius CovalentRadius    Color
*/
	{"H",  "hydrogen",       1.00797, 	1.54, 	0.32, 	{255, 255, 255}},	/*Atom 1*/
	{"He", "helium",          4.0026, 	1.12, 	0.93,	{255, 255, 161}},	/*Atom 2*/
	{"Li", "lithium",          6.937, 	0.68, 	1.23,	{187, 255, 255}},	/*Atom 3*/
	{"Be", "berylium",        9.0122, 	0.44, 	0.90,	{  0, 255, 184}},	/*Atom 4*/
	{"B",  "boron",           10.811, 	0.35, 	0.82,	{128,  58,  43}},	/*Atom 5*/
	{"C",  "carbon",         12.0112, 	2.60, 	0.77,	{  0, 200,   0}},	/*Atom 6*/
	{"N",  "nitrogen",       14.0067, 	1.71, 	0.75,	{  0, 128, 255}},	/*Atom 7*/
	{"O",  "oxygen",         15.9994, 	1.32, 	0.73,	{255,   0,   0}},	/*Atom 8*/
	{"F",  "fluorine",       18.9984, 	1.33, 	0.72,	{255,   0,   0}},	/*Atom 9*/
	{"Ne", "neon",            20.183, 	1.12, 	0.71,	{255,  69, 213}},	/*Atom 10*/
	{"Na", "sodium",         22.9898, 	0.97, 	1.54,	{255, 128,   0}},	/*Atom 11*/
	{"Mg", "magnesium",       24.312, 	0.82, 	1.36,	{128,  70,  76}},	/*Atom 12*/
	{"Al", "aluminum",       26.9815, 	0.51, 	1.18,	{194, 247, 255}},	/*Atom 13*/
	{"Si", "silicon",         28.086, 	2.71, 	1.11,	{255, 280, 120}},	/*Atom 14*/
	{"P",  "phosphorus",     30.9738, 	2.12, 	1.06,	{255, 255,  80}},	/*Atom 15*/
	{"S",  "sulphur",        32.9738, 	1.84, 	1.02,	{255, 191,   0}},	/*Atom 16*/
	{"Cl", "chlorine",        35.453, 	1.81, 	0.99,	{  0, 255, 255}},	/*Atom 17*/
	{"Ar", "argon",           39.948, 	1.54, 	0.98,	{255,  51,  55}},	/*Atom 18*/
	{"K",  "potassium",       39.102, 	1.33, 	2.03,	{200, 100,   0}},	/*Atom 19*/
	{"Ca", "calcium",          40.08, 	1.18, 	1.74,	{255,   0,   0}},	/*Atom 20*/
	{"Sc", "scandium",        44.956,	0.732, 	1.44,	{255,   0,   0}},	/*Atom 21*/
	{"Ti", "titanium",          47.9, 	0.96, 	1.32,	{255,   0,   0}},	/*Atom 22*/
	{"V",  "vanadium",        50.942, 	0.88, 	1.22,	{255,   0,   0}},	/*Atom 23*/
	{"Cr", "chromium",        51.996, 	0.81, 	1.18,	{255,   0,   0}},	/*Atom 24*/
	{"Mn", "manganese",        54.93, 	0.80, 	1.17,	{255,   0,   0}},	/*Atom 25*/
	{"Fe", "iron",            55.847, 	0.74, 	1.17,	{255,   0,   0}},	/*Atom 26*/
	{"Co", "cobalt",         58.9332, 	0.72, 	1.16,	{ 44,   0, 255}},	/*Atom 27*/
	{"Ni", "nickel",           58.71, 	0.69, 	1.15,	{255,   0,   0}},	/*Atom 28*/
	{"Cu", "copper",          63.546, 	0.96, 	1.17,	{255,   0,   0}},	/*Atom 29*/
	{"Zn", "zinc",             65.37, 	0.88, 	1.25,	{255,   0,   0}},	/*Atom 30*/
	{"Ga", "gallium",          69.72, 	0.81, 	1.26,	{255,   0,   0}},	/*Atom 31*/
	{"Ge", "germanium",        72.59, 	2.72, 	1.22,	{255,   0,   0}},	/*Atom 32*/
	{"As", "arsenic",        74.9216, 	2.22, 	1.20,	{255,   0,   0}},	/*Atom 33*/
	{"Se", "selenium",         78.96, 	1.91, 	1.16,	{255,   0,   0}},	/*Atom 34*/
	{"Br", "bromine",         79.909, 	1.96, 	1.14,	{255,   0,   0}},	/*Atom 35*/
	{"Kr", "krypton",           83.8, 	1.12, 	1.12,	{255,   0,   0}},	/*Atom 36*/
	{"Rb", "rubidium",         85.47, 	1.47, 	2.16,	{255,   0,   0}},	/*Atom 37*/
	{"Sr", "strontium",        87.62, 	1.12, 	1.91,	{255,   0,   0}},	/*Atom 38*/
	{"Y",  "yttrium",         88.905,	0.893, 	1.62,	{255,   0,   0}},	/*Atom 39*/
	{"Zr", "zirconium",        91.22, 	1.09, 	1.45,	{255,   0,   0}},	/*Atom 40*/
	{"Nb", "niobium",         92.906, 	1.00, 	1.34,	{255,   0,   0}},	/*Atom 41*/
	{"Mo", "molybdenum",       95.94, 	0.93, 	1.30,	{255,   0,   0}},	/*Atom 42*/
	{"Tc", "technetium",          97,	0.979, 	1.27,	{255,   0,   0}},	/*Atom 43*/
	{"Ru", "ruthenium",       101.07, 	0.67, 	1.25,	{255,   0,   0}},	/*Atom 44*/
	{"Rh", "rhodium",        102.905, 	0.68, 	1.25,	{255,   0,   0}},	/*Atom 45*/
	{"Pd", "palladium",        106.4, 	0.80, 	1.28,	{255,   0,   0}},	/*Atom 46*/
	{"Ag", "silver",         107.868, 	1.26, 	1.34,	{255,   0,   0}},	/*Atom 47*/
	{"Cd", "cadmium",          112.4, 	1.14, 	1.48,	{255,   0,   0}},	/*Atom 48*/
	{"In", "indium",          114.82, 	0.81, 	1.44,	{255,   0,   0}},	/*Atom 49*/
	{"Sn", "tin",             118.69, 	2.94, 	1.41,	{ 76, 156, 256}},	/*Atom 50*/
	{"Sb", "antimony",        121.75, 	2.45, 	1.40,	{255,   0,   0}},	/*Atom 51*/
	{"Te", "tellurium",        127.6, 	2.11, 	1.36,	{255,   0,   0}},	/*Atom 52*/
	{"I",  "iodine",         136.904, 	2.20, 	1.33,	{255,   0,   0}},	/*Atom 53*/
	{"Xe", "xenon",            131.3, 	1.12, 	1.31,	{255,   0,   0}},	/*Atom 54*/
	{"Cs", "cesium",         132.905, 	1.67, 	2.35,	{255,   0,   0}},	/*Atom 55*/
	{"Ba", "barium",          137.34, 	1.53, 	1.98,	{255,   0,   0}},	/*Atom 56*/
	{"La", "lanthanum",       138.91, 	1.39, 	1.69,	{255,   0,   0}},	/*Atom 57*/
	{"Ce", "cerium",          140.12, 	1.27, 	1.65,	{255,   0,   0}},	/*Atom 58*/
	{"Pr", "praseodymium",   141.907,	1.013, 	1.65,	{255,   0,   0}},	/*Atom 59*/
	{"Nd", "neodymium",       144.24,	0.995, 	1.64,	{255,   0,   0}},	/*Atom 60*/
	{"Pm", "promethium",         145,	0.979, 	1.63,	{255,   0,   0}},	/*Atom 61*/
	{"Sm", "samarium",        150.35,	0.964, 	1.62,	{255,   0,   0}},	/*Atom 62*/
	{"Eu", "europium",        151.96,	0.950, 	1.85,	{255,   0,   0}},	/*Atom 63*/
	{"Gd", "gadolinium",      157.25,	0.938, 	1.61,	{255,   0,   0}},	/*Atom 64*/
	{"Tb", "terbium",        158.924,	0.923, 	1.59,	{255,   0,   0}},	/*Atom 65*/
	{"Dy", "dysprosium",       162.5,	0.908, 	1.59,	{255,   0,   0}},	/*Atom 66*/
	{"Ho", "holmium",         164.93,	0.894, 	1.58,	{255,   0,   0}},	/*Atom 67*/
	{"Er", "erbium",          167.28,	0.881, 	1.57,	{255,   0,   0}},	/*Atom 68*/
	{"Tm", "thulium",        168.934, 	0.87, 	1.56,	{255,   0,   0}},	/*Atom 69*/
	{"Yb", "ytterbium",       173.04, 	0.93, 	1.74,	{255,   0,   0}},	/*Atom 70*/
	{"Lu", "lutatium",        174.97, 	0.85, 	1.56,	{255,   0,   0}},	/*Atom 71*/
	{"Hf", "hafnium",         178.49, 	0.78, 	1.44,	{255,   0,   0}},	/*Atom 72*/
	{"Ta", "tantalum",       180.948, 	0.68, 	1.34,	{255,   0,   0}},	/*Atom 73*/
	{"W",  "tungsten",        193.95, 	0.70, 	1.30,	{255,   0,   0}},	/*Atom 74*/
	{"Re", "rhenium",          186.2, 	0.72, 	1.28,	{255,   0,   0}},	/*Atom 75*/
	{"Os", "osmium",           190.2, 	0.88, 	1.26,	{255,   0,   0}},	/*Atom 76*/
	{"Ir", "iridium",          192.2, 	0.68, 	1.27,	{255,   0,   0}},	/*Atom 77*/
	{"Pt", "platinum",       195.096, 	0.80, 	1.30,	{255,   0,   0}},	/*Atom 78*/
	{"Au", "gold",           196.967, 	1.37, 	1.34,	{255, 179,   0}},	/*Atom 79*/
	{"Hg", "mercury",         200.59, 	1.27, 	1.49,	{255,   0,   0}},	/*Atom 80*/
	{"Tl", "thallium",        204.37, 	1.47, 	1.48,	{255,   0,   0}},	/*Atom 81*/
	{"Pb", "lead",            207.19, 	1.20, 	1.47,	{ 15,  31,  51}},	/*Atom 82*/
	{"Bi", "bismuth",         208.98, 	0.98, 	1.46,	{255,   0,   0}},	/*Atom 83*/
	{"Po", "polonium",           210, 	0.67, 	1.46,	{255,   0,   0}},	/*Atom 84*/
	{"At", "astatine",           210, 	0.62, 	1.45,	{255,   0,   0}},	/*Atom 85*/
	{"Rn", "radon",              222, 	1.12,   MISRAD,	{255,   0,   0}},	/*Atom 86*/
	{"Fr", "francium",           223, 	1.80, 	MISRAD,	{255,   0,   0}},	/*Atom 87*/
	{"Ra", "radium",             226, 	1.43, 	MISRAD,	{255,   0,   0}},	/*Atom 88*/
	{"Ac", "actinium",           227, 	1.18, 	MISRAD,	{255,   0,   0}},	/*Atom 89*/
	{"Th", "thorium",         232.08, 	1.02, 	1.65,	{255,   0,   0}},	/*Atom 90*/
	{"Pa", "protactinium",       231, 	1.13, 	MISRAD,	{255,   0,   0}},	/*Atom 91*/
	{"U",  "uranium",         238.03, 	0.97, 	1.42,	{255,   0,   0}},	/*Atom 92*/
	{"Np", "neptunium",          237, 	1.10, 	MISRAD,	{255,   0,   0}},	/*Atom 93*/
	{"Pu", "plutonium",          244, 	1.08, 	MISRAD,	{255,   0,   0}},	/*Atom 94*/
	{"Am", "americium",          243, 	1.07, 	MISRAD,	{255,   0,   0}},	/*Atom 95*/
	{"Cm", "curium",             247, 	1.08, 	MISRAD,	{255,   0,   0}},	/*Atom 96*/
	{"Bk", "berkelium",          247, 	1.08,	MISRAD, 	{255,   0,   0}},	/*Atom 97*/
	{"Cf", "californium",        251, 	1.08, 	MISRAD,	{255,   0,   0}},	/*Atom 98*/
	{"Es", "einsteinium",        254, 	1.08, 	MISRAD,	{255,   0,   0}},	/*Atom 99*/
	{"Fm", "fermium",            257, 	1.08, 	MISRAD,	{255,   0,   0}},	/*Atom 100*/
	{"Md", "mendelevium",        256, 	1.08, 	MISRAD,	{255,   0,   0}},	/*Atom 101*/
	{"No", "nobelium",           254, 	1.08, 	MISRAD,	{255,   0,   0}},	/*Atom 102*/
	{"Lw", "lawrencium",         257, 	1.08, 	MISRAD,	{255,   0,   0}},	/*Atom 103*/
    };
